typedef struct {
	signed short I;
	signed short Q;
} complexInt;

typedef struct {
	double I;
	double Q;
} complexDbl;

typedef struct {
	float I;
	float Q;
} complexFlt;

//Structs for WARPNet Server <-> iPhone Controller communications
typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char modOrder;
	unsigned char txPower;
	unsigned char coding;
	unsigned char channel;
} warpnetControl;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char partnerID;
	struct timeval lastUpdate;
	unsigned int goodPackets;
	unsigned int partnerBadPackets;
	unsigned int rxBytes;
	unsigned int txBytes;
	unsigned int otherBadPackets;
	float throughput;
} warpnetObserve;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char trafficMode; //0-bridge, 1-traffic source, 2-traffic sync, 3-src/sync
	unsigned char reserved0;
	unsigned int txInterval; //This changes the time
	unsigned short txPacketLen;
	unsigned short reserved1;
} warpnetTraffic;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned short cmdID;
	unsigned int cmdParam;
} warpnetCommand;

typedef struct {
	unsigned char structID;
	unsigned char numControllers;
	unsigned char numWARPnodes;
} warpServer_status __attribute__((__aligned__(8)));

//Structs for WARPNet Server <-> WARP Node communication
typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char partnerID;
	unsigned char reserved0;
	unsigned int goodPackets;
	unsigned int partnerBadPackets;
	unsigned int otherBadPackets;
	unsigned int rxBytes;
	unsigned int txBytes;
	unsigned long long time;
} warpnodeStats;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;	
} warpnodeID;

//Struct containing control parameters for the MAC/PHY
typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char modOrder;
	unsigned char txPower;
	unsigned char coding;
	unsigned char channel;
} warpnodeControl;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char trafficMode; //0-bridge, 1-traffic source, 2-traffic sync, 3-src/sync
	unsigned char reserved0;
	unsigned int txInterval; //This changes the time
	unsigned short txPacketLen;
	unsigned short reserved1;
} warpnodeTraffic;

//Struct containing a command for a node
typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned short cmdID;
	unsigned int cmdParam;
} warpnodeCommand;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char partnerID;
	unsigned char antennaID;
	unsigned char agc_RF;
	unsigned char agc_BB;
	unsigned char pktStatus;
	unsigned char reserved0;
	//64*4 bytes must follow in the actual Ethernet packet!
} warpnodeChanEst;

typedef struct {
	unsigned char structType;
	unsigned char nodeID;
	unsigned char partnerID;
	unsigned char antennaID;
	unsigned char agc_RF;
	unsigned char agc_BB;
	unsigned char pktStatus;
	
//	complexDbl chanEsts[64];
	complexFlt chanEsts[64];
} chanEstFlt;

//Struct representing raw Ethernet header + packet contents
typedef struct {
	//Standard 14-byte Ethernet header (dest/src MAC addresses + EtherType)
	unsigned char dstAddr[6];
	unsigned char srcAddr[6];
	unsigned short ethType;
	
	//total number of bytes in pkt, including this header
	unsigned short pktLength;
	
	//number of warpnet structs that follow
	unsigned char numStructs;
	
	//Sequence number to match packets to ACKs
	unsigned char seqNum;
} warpnetEthernetPktHeader;

//Struct Types for iPhone Controller <-> Server Packets
#define STRUCTID_SERVERSTATUS 0x30
#define STRUCTID_OBSERVE 0x31
#define STRUCTID_CONTROL 0x32
#define STRUCTID_COMMAND 0x33
#define STRUCTID_TRAFFIC 0x34

//Struct Types for WARP Node <-> Server Packets
#define STRUCTID_NODESTATS 0x40
#define STRUCTID_NODEID 0x41
#define STRUCTID_NODECONTROL 0x42
#define STRUCTID_NODECOMMAND 0x43
#define STRUCTID_CHANEST 0x44
#define STRUCTID_NODETRAFFIC 0x45
#define CHAN_EST_NUMBYTES 256

//Node command codes
#define NODECMD_RESETSTATS 0x61
#define NODECMD_REBOOT 0x62
#define NODECMD_NODEACK 0x63

//Initial node identification stuff
#define MAX_NODEID_ATTEMPTS 100
#define NODEID_UNKNOWN 0xFF
#define OBSERVE_NODATA 0xFFFFFFFF

#define WARPNET_ETHTYPE_SVR2NODE 0x9090
#define WARPNET_ETHTYPE_NODE2SVR 0x9191

//Compile-time limits on numbers of nodes and controllers
#define MAX_NUMCONTROLLERS 16
#define MAX_NUMWARPNODES 16
#define MAX_NODETXTRIES 32
#define NODETX_TRYDELAY 65000

//Parameters for advertised Bonjour service
#define SERVICE_TYPE "_warpnet._tcp"
//#define SERVICE_NAME "server X"
#define SERVICE_PORT 9090

//Maximum size of network buffers
#define MAX_TCPDATA 65535
#define MAX_ETHPKTSIZE 1550

#define XPAR_PPC405_CORE_CLOCK_FREQ_HZ 240000000

#define ntohll(x) (((long long)(ntohl((int)((x << 32) >> 32))) << 32) | (unsigned int)ntohl(((int)(x >> 32)))) //By Runner

//Uncomment this define to emulate WARP nodes statistics updates
//#define EMULATE_WARPNODESTATS 1


void PrintStruct(void * theStruct);
void ProcessWARPnodeMessage(u_char *args, const struct pcap_pkthdr *header, const u_char *packet);
void *pcapLoopThread(void* theArgs);
int SetupPCAP();
void RegisterBonjourService();
int UpdateController(int sock, void* theStruct);
int ProcessControllerMessage(unsigned char * RxMsg, int numRxBytes, int controller_sock);
void* ControlConnectionHandler(void* args);
int SetupControlServer(void);
void IdenfityLocalWARPNode();
void* VerifyNodeControlState();
